@@ -1,7 +1,7 @@
 /*
  * kafkacat - Apache Kafka consumer and producer
  *
- * Copyright (c) 2014, Magnus Edenhill
+ * Copyright (c) 2014-2019, Magnus Edenhill
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -71,12 +71,12 @@ static struct stats {
 } stats;
 
 
-/* Partition's at EOF state array */
-int *part_eof = NULL;
-/* Number of partitions that has reached EOF */
-int part_eof_cnt = 0;
-/* Threshold level (partitions at EOF) before exiting */
-int part_eof_thres = 0;
+/* Partition's stopped state array */
+int *part_stop = NULL;
+/* Number of partitions that are stopped */
+int part_stop_cnt = 0;
+/* Threshold level (partitions stopped) before exiting */
+int part_stop_thres = 0;
 
 
 
@@ -182,8 +182,6 @@ static void produce (void *buf, size_t len,
                         break;
                 }
 
-                err = rd_kafka_last_error();
-
                 if (err != RD_KAFKA_RESP_ERR__QUEUE_FULL)
                         KC_FATAL("Failed to produce message (%zd bytes): %s",
                               len, rd_kafka_err2str(err));
@@ -294,7 +292,7 @@ static void producer_run (FILE *fp, char **paths, int pathcnt) {
                                      errstr, sizeof(errstr))))
                 KC_FATAL("Failed to create producer: %s", errstr);
 
-		if (!conf.debug && conf.verbosity == 0)
+        if (!conf.debug && conf.verbosity == 0)
                 rd_kafka_set_log_level(conf.rk, 0);
 
         /* Create topic */
@@ -431,6 +429,17 @@ static void producer_run (FILE *fp, char **paths, int pathcnt) {
                 conf.exitcode = 1;
 }
 
+static void stop_partition (rd_kafka_message_t *rkmessage) {
+        if (!part_stop[rkmessage->partition]) {
+                /* Stop consuming this partition */
+                rd_kafka_consume_stop(rkmessage->rkt,
+                                      rkmessage->partition);
+                part_stop[rkmessage->partition] = 1;
+                part_stop_cnt++;
+                if (part_stop_cnt >= part_stop_thres)
+                        conf.run = 0;
+        }
+}
 
 static void handle_partition_eof (rd_kafka_message_t *rkmessage) {
 
@@ -443,15 +452,7 @@ static void handle_partition_eof (rd_kafka_message_t *rkmessage) {
                                       rkmessage->offset == 0 ?
                                       0 : rkmessage->offset-1);
                 if (conf.exit_eof) {
-                        if (!part_eof[rkmessage->partition]) {
-                                /* Stop consuming this partition */
-                                rd_kafka_consume_stop(rkmessage->rkt,
-                                                      rkmessage->partition);
-                                part_eof[rkmessage->partition] = 1;
-                                part_eof_cnt++;
-                                if (part_eof_cnt >= part_eof_thres)
-                                        conf.run = 0;
-                        }
+                        stop_partition(rkmessage);
                 }
 
         } else if (conf.mode == 'G') {
@@ -494,6 +495,20 @@ static void consume_cb (rd_kafka_message_t *rkmessage, void *opaque) {
 
         }
 
+        if (conf.stopts) {
+                int64_t ts = rd_kafka_message_timestamp(rkmessage, NULL);
+                if (ts >= conf.stopts) {
+                        stop_partition(rkmessage);
+                        KC_INFO(1, "Reached stop timestamp for topic %s [%"PRId32"] "
+                             "at offset %"PRId64"%s\n",
+                             rd_kafka_topic_name(rkmessage->rkt),
+                             rkmessage->partition,
+                             rkmessage->offset,
+                             !conf.run ? ": exiting" : "");
+                        return;                       
+                }
+        }
+
         /* Print message */
         fmt_msg_output(fp, rkmessage);
 
@@ -628,6 +643,44 @@ static void kafkaconsumer_run (FILE *fp, char *const *topics, int topic_cnt) {
 }
 #endif
 
+/**
+ * Get offsets from conf.startts for consumer_run
+ */
+static int64_t *get_offsets(rd_kafka_metadata_topic_t *topic) {
+        int i;
+        int64_t *offsets;
+        rd_kafka_resp_err_t err;
+        rd_kafka_topic_partition_list_t *rktparlistp =
+                                           rd_kafka_topic_partition_list_new(1);
+
+        for (i = 0 ; i < topic->partition_cnt ; i++) {
+                int32_t partition = topic->partitions[i].id;
+
+                /* If -p <part> was specified: skip unwanted partitions */
+                if (conf.partition != RD_KAFKA_PARTITION_UA &&
+                    conf.partition != partition)
+                        continue;
+
+                rd_kafka_topic_partition_list_add(rktparlistp,
+                    rd_kafka_topic_name(conf.rkt),
+                    partition)->offset = conf.startts;
+
+                if (conf.partition != RD_KAFKA_PARTITION_UA)
+                        break;
+        }
+        err = rd_kafka_offsets_for_times(conf.rk, rktparlistp, 10*1000);
+        if (err)
+                KC_FATAL("offsets_for_times failed: %s", rd_kafka_err2str(err));
+
+        offsets = calloc(sizeof(int64_t), topic->partition_cnt);
+        for (i = 0 ; i < rktparlistp->cnt ; i++) {
+                const rd_kafka_topic_partition_t *p = &rktparlistp->elems[i];
+                offsets[p->partition] = p->offset;
+        }
+        rd_kafka_topic_partition_list_destroy(rktparlistp);
+
+        return offsets;
+}
 
 /**
  * Run consumer, consuming messages from Kafka and writing to 'fp'.
@@ -637,12 +690,13 @@ static void consumer_run (FILE *fp) {
         rd_kafka_resp_err_t err;
         const rd_kafka_metadata_t *metadata;
         int i;
+        int64_t *offsets = NULL;
         rd_kafka_queue_t *rkqu;
 
         /* Create consumer */
         if (!(conf.rk = rd_kafka_new(RD_KAFKA_CONSUMER, conf.rk_conf,
                                      errstr, sizeof(errstr))))
-                KC_FATAL("Failed to create producer: %s", errstr);
+                KC_FATAL("Failed to create consumer: %s", errstr);
 
         if (!conf.debug && conf.verbosity == 0)
                 rd_kafka_set_log_level(conf.rk, 0);
@@ -685,16 +739,22 @@ static void consumer_run (FILE *fp) {
 
         /* If Exit-at-EOF is enabled, set up array to track EOF
          * state for each partition. */
-        if (conf.exit_eof) {
-                part_eof = calloc(sizeof(*part_eof),
+        if (conf.exit_eof || conf.stopts) {
+                part_stop = calloc(sizeof(*part_stop),
                                   metadata->topics[0].partition_cnt);
 
                 if (conf.partition != RD_KAFKA_PARTITION_UA)
-                        part_eof_thres = 1;
+                        part_stop_thres = 1;
                 else
-                        part_eof_thres = metadata->topics[0].partition_cnt;
+                        part_stop_thres = metadata->topics[0].partition_cnt;
         }
 
+#if RD_KAFKA_VERSION >= 0x00090300
+        if (conf.startts) {
+            offsets = get_offsets(&metadata->topics[0]);
+        }
+#endif
+
         /* Create a shared queue that combines messages from
          * all wanted partitions. */
         rkqu = rd_kafka_queue_new(conf.rk);
@@ -710,7 +770,8 @@ static void consumer_run (FILE *fp) {
 
                 /* Start consumer for this partition */
                 if (rd_kafka_consume_start_queue(conf.rkt, partition,
-                                                 conf.offset, rkqu) == -1)
+                                                 offsets ? offsets[i] : conf.offset,
+                                                 rkqu) == -1)
                         KC_FATAL("Failed to start consuming "
                                  "topic %s [%"PRId32"]: %s",
                                  conf.topic, partition,
@@ -719,6 +780,7 @@ static void consumer_run (FILE *fp) {
                 if (conf.partition != RD_KAFKA_PARTITION_UA)
                         break;
         }
+        free(offsets);
 
         if (conf.partition != RD_KAFKA_PARTITION_UA &&
             i == metadata->topics[0].partition_cnt)
@@ -748,7 +810,7 @@ static void consumer_run (FILE *fp) {
                         continue;
 
 		/* Dont stop already stopped partitions */
-		if (!part_eof || !part_eof[partition])
+		if (!part_stop || !part_stop[partition])
 			rd_kafka_consume_stop(conf.rkt, partition);
 
                 rd_kafka_consume_stop(conf.rkt, partition);
@@ -916,15 +978,18 @@ static void RD_NORETURN usage (const char *argv0, int exitcode,
         fprintf(out,
                 "kafkacat - Apache Kafka producer and consumer tool\n"
                 "https://github.com/edenhill/kafkacat\n"
-                "Copyright (c) 2014-2017, Magnus Edenhill\n"
-                "Version %s%s (librdkafka %s builtin.features=%s)\n"
+                "Copyright (c) 2014-2019, Magnus Edenhill\n"
+                "Version %s (%slibrdkafka %s builtin.features=%s)\n"
                 "\n",
                 KAFKACAT_VERSION,
+                ""
 #if ENABLE_JSON
-                " (JSON)",
-#else
-                "",
+                "JSON, "
 #endif
+#if ENABLE_AVRO
+                "Avro, "
+#endif
+                ,
                 rd_kafka_version_str(), features
                 );
 
@@ -959,6 +1024,10 @@ static void RD_NORETURN usage (const char *argv0, int exitcode,
                 "  -X prop=val        Set librdkafka configuration property.\n"
                 "                     Properties prefixed with \"topic.\" are\n"
                 "                     applied as topic properties.\n"
+#if ENABLE_AVRO
+                "  -X schema.registry.prop=val Set libserdes configuration property "
+                "for the Avro/Schema-Registry client.\n"
+#endif
                 "  -X dump            Dump configuration and exit.\n"
                 "  -d <dbg1,...>      Enable librdkafka debugging:\n"
                 "                     " RD_KAFKA_DEBUG_CONTEXTS "\n"
@@ -994,12 +1063,46 @@ static void RD_NORETURN usage (const char *argv0, int exitcode,
                 "                     beginning | end | stored |\n"
                 "                     <value>  (absolute offset) |\n"
                 "                     -<value> (relative offset from end)\n"
+#if RD_KAFKA_VERSION >= 0x00090300
+                "                     s@<value> (timestamp in ms to start at)\n"
+                "                     e@<value> (timestamp in ms to stop at "
+                "(not included))\n"
+#endif
                 "  -e                 Exit successfully when last message "
                 "received\n"
                 "  -f <fmt..>         Output formatting string, see below.\n"
                 "                     Takes precedence over -D and -K.\n"
 #if ENABLE_JSON
                 "  -J                 Output with JSON envelope\n"
+#endif
+                "  -s key=<serdes>    Deserialize non-NULL keys using <serdes>.\n"
+                "  -s value=<serdes>  Deserialize non-NULL values using <serdes>.\n"
+                "  -s <serdes>        Deserialize non-NULL keys and values using <serdes>.\n"
+                "                     Available deserializers (<serdes>):\n"
+                "                       <pack-str> - A combination of:\n"
+                "                                    <: little-endian,\n"
+                "                                    >: big-endian (recommended),\n"
+                "                                    b: signed 8-bit integer\n"
+                "                                    B: unsigned 8-bit integer\n"
+                "                                    h: signed 16-bit integer\n"
+                "                                    H: unsigned 16-bit integer\n"
+                "                                    i: signed 32-bit integer\n"
+                "                                    I: unsigned 32-bit integer\n"
+                "                                    q: signed 64-bit integer\n"
+                "                                    Q: unsigned 64-bit integer\n"
+                "                                    c: ASCII character\n"
+                "                                    s: remaining data is string\n"
+                "                                    $: match end-of-input (no more bytes remaining or a parse error is raised).\n"
+                "                                       Not including this token skips any\n"
+                "                                       remaining data after the pack-str is\n"
+                "                                       exhausted.\n"
+#if ENABLE_AVRO
+                "                       avro       - Avro-formatted with schema in Schema-Registry (requires -r)\n"
+                "                     E.g.: -s key=i -s value=avro - key is 32-bit integer, value is Avro.\n"
+                "                       or: -s avro - both key and value are Avro-serialized\n"
+#endif
+#if ENABLE_AVRO
+                "  -r <url>           Schema registry URL (requires avro deserializer to be used with -s)\n"
 #endif
                 "  -D <delim>         Delimiter to separate messages on output\n"
                 "  -K <delim>         Print message keys prefixing the message\n"
@@ -1007,8 +1110,9 @@ static void RD_NORETURN usage (const char *argv0, int exitcode,
                 "  -O                 Print message offset using -K delimiter\n"
                 "  -c <cnt>           Exit after consuming this number "
                 "of messages\n"
-                "  -Z                 Print NULL messages and keys as \"%s\""
-                "(instead of empty)\n"
+                "  -Z                 Print NULL values and keys as \"%s\""
+                "instead of empty.\n"
+                "                     For JSON (-J) the nullstr is always null.\n"
                 "  -u                 Unbuffered output\n"
                 "\n"
                 "Metadata options (-L):\n"
@@ -1044,6 +1148,15 @@ static void RD_NORETURN usage (const char *argv0, int exitcode,
                 " Example:\n"
                 "  -f 'Topic %%t [%%p] at offset %%o: key %%k: %%s\\n'\n"
                 "\n"
+                "JSON message envelope (on one line) when consuming with -J:\n"
+                " { \"topic\": str, \"partition\": int, \"offset\": int,\n"
+                "   \"tstype\": \"create|logappend|unknown\", \"ts\": int, "
+                "// timestamp in milliseconds since epoch\n"
+                "   \"headers\": { \"<name>\": str, .. }, // optional\n"
+                "   \"key\": str|json, \"payload\": str|json,\n"
+                "   \"key_error\": str, \"payload_error\": str } //optional\n"
+                " (note: key_error and payload_error are only included if "
+                "deserialization failed)\n"
                 "\n"
                 "Consumer mode (writes messages to stdout):\n"
                 "  kafkacat -b <broker> -t <topic> -p <partition>\n"
@@ -1199,6 +1312,30 @@ static void conf_dump (void) {
 static int try_conf_set (const char *name, const char *val,
                          char *errstr, size_t errstr_size) {
         rd_kafka_conf_res_t res = RD_KAFKA_CONF_UNKNOWN;
+        size_t srlen = strlen("schema.registry.");
+
+        /* Pass schema.registry. config to the serdes */
+        if (!strncmp(name, "schema.registry.", srlen)) {
+#if ENABLE_AVRO
+                serdes_err_t serr;
+
+                if (!conf.srconf)
+                        conf.srconf = serdes_conf_new(NULL, 0, NULL);
+
+                if (!strcmp(name, "schema.registry.url"))
+                        srlen = 0;
+
+                serr = serdes_conf_set(conf.srconf, name+srlen, val,
+                                       errstr, errstr_size);
+                return serr == SERDES_ERR_OK ? 0 : -1;
+#else
+                snprintf(errstr, errstr_size,
+                         "This build of kafkacat lacks "
+                         "Avro/Schema-Registry support");
+                return -1;
+#endif
+        }
+
 
         /* Try "topic." prefixed properties on topic
          * conf first, and then fall through to global if
@@ -1430,13 +1567,11 @@ static void argparse (int argc, char **argv,
         char tmp_fmt[64];
         int do_conf_dump = 0;
         int conf_files_read = 0;
+        int i;
 
         while ((opt = getopt(argc, argv,
-                             "PCG:LQt:p:b:z:o:eED:K:k:H:Od:qvF:X:c:Tuf:ZlVh"
-#if ENABLE_JSON
-                             "J"
-#endif
-                        )) != -1) {
+                             ":PCG:LQt:p:b:z:o:eED:K:k:H:Od:qvF:X:c:Tuf:ZlVh"
+                             "s:r:J")) != -1) {
                 switch (opt) {
                 case 'P':
                 case 'C':
@@ -1459,7 +1594,7 @@ static void argparse (int argc, char **argv,
                                 if (!*rktparlistp)
                                         *rktparlistp = rd_kafka_topic_partition_list_new(1);
                                 add_topparoff("-t", *rktparlistp, optarg);
-
+                                conf.flags |= CONF_F_APIVERREQ;
                         } else
                                 conf.topic = optarg;
 
@@ -1485,6 +1620,15 @@ static void argparse (int argc, char **argv,
                                 conf.offset = RD_KAFKA_OFFSET_BEGINNING;
                         else if (!strcmp(optarg, "stored"))
                                 conf.offset = RD_KAFKA_OFFSET_STORED;
+#if RD_KAFKA_VERSION >= 0x00090300
+                        else if (!strncmp(optarg, "s@", 2)) {
+                                conf.startts = strtoll(optarg+2, NULL, 10);
+                                conf.flags |= CONF_F_APIVERREQ;
+                        } else if (!strncmp(optarg, "e@", 2)) {
+                                conf.stopts = strtoll(optarg+2, NULL, 10);
+                                conf.flags |= CONF_F_APIVERREQ;
+                        }
+#endif
                         else {
                                 conf.offset = strtoll(optarg, NULL, 10);
                                 if (conf.offset < 0)
@@ -1500,11 +1644,50 @@ static void argparse (int argc, char **argv,
                 case 'f':
                         fmt = optarg;
                         break;
-#if ENABLE_JSON
                 case 'J':
+#if ENABLE_JSON
                         conf.flags |= CONF_F_FMT_JSON;
-                        break;
+#else
+                        KC_FATAL("This build of kafkacat lacks JSON support");
 #endif
+                        break;
+
+                case 's':
+                {
+                        int field = -1;
+                        const char *t = optarg;
+
+                        if (!strncmp(t, "key=", strlen("key="))) {
+                                t += strlen("key=");
+                                field = KC_MSG_FIELD_KEY;
+                        } else if (!strncmp(t, "value=", strlen("value="))) {
+                                t += strlen("value=");
+                                field = KC_MSG_FIELD_VALUE;
+                        }
+
+                        if (field == -1 || field == KC_MSG_FIELD_KEY) {
+                                if (strcmp(t, "avro"))
+                                        pack_check("key", t);
+                                conf.pack[KC_MSG_FIELD_KEY] = t;
+                        }
+
+                        if (field == -1 || field == KC_MSG_FIELD_VALUE) {
+                                if (strcmp(t, "avro"))
+                                        pack_check("value", t);
+                                conf.pack[KC_MSG_FIELD_VALUE] = t;
+                        }
+                }
+                break;
+                case 'r':
+#if ENABLE_AVRO
+                        if (!*optarg)
+                                KC_FATAL("-s url must not be empty");
+                        conf.schema_registry_url = optarg;
+#else
+                        KC_FATAL("This build of kafkacat lacks "
+                                 "Avro/Schema-Registry support");
+#endif
+                        break;
                 case 'D':
                         delim = optarg;
                         break;
@@ -1606,6 +1789,7 @@ static void argparse (int argc, char **argv,
                 }
         }
 
+
         if (conf_files_read == 0) {
                 const char *cpath = kc_getenv("KAFKACAT_CONFIG");
                 if (cpath) {
@@ -1640,7 +1824,9 @@ static void argparse (int argc, char **argv,
         if (!strchr("GLQ", conf.mode) && !conf.topic)
                 usage(argv[0], 1, "-t <topic> missing", 0);
         else if (conf.mode == 'Q' && !*rktparlistp)
-                usage(argv[0], 1, "-t <topic>:<partition>:<offset_or_timestamp> missing", 0);
+                usage(argv[0], 1,
+                      "-t <topic>:<partition>:<offset_or_timestamp> missing",
+                      0);
 
         if (conf.brokers &&
             rd_kafka_conf_set(conf.rk_conf, "metadata.broker.list",
@@ -1652,6 +1838,80 @@ static void argparse (int argc, char **argv,
 
         fmt_init();
 
+        /*
+         * Verify serdes
+         */
+        for (i = 0 ; i < KC_MSG_FIELD_CNT ; i++) {
+                if (!conf.pack[i])
+                        continue;
+
+                if (!strchr("GC", conf.mode))
+                        KC_FATAL("-s serdes only available in the consumer");
+
+                if (conf.pack[i] && !strcmp(conf.pack[i], "avro")) {
+#if !ENABLE_AVRO
+                        KC_FATAL("This build of kafkacat lacks "
+                                 "Avro/Schema-Registry support");
+#endif
+#if ENABLE_JSON
+                        /* Avro is decoded to JSON which needs to be
+                         * written verbatim to the JSON envelope when
+                         * using -J: libyajl does not support this,
+                         * but my own fork of yajl does. */
+                        if (conf.flags & CONF_F_FMT_JSON &&
+                            !json_can_emit_verbatim())
+                                KC_FATAL("This build of kafkacat lacks "
+                                         "support for emitting "
+                                         "JSON-formatted "
+                                         "message keys and values: "
+                                         "try without -J or build "
+                                         "kafkacat with yajl from "
+                                         "https://github.com/edenhill/yajl");
+#endif
+
+                        if (i == KC_MSG_FIELD_VALUE)
+                                conf.flags |= CONF_F_FMT_AVRO_VALUE;
+                        else if (i == KC_MSG_FIELD_KEY)
+                                conf.flags |= CONF_F_FMT_AVRO_KEY;
+                        continue;
+                }
+        }
+
+
+        /*
+         * Verify and initialize Avro/SR
+         */
+#if ENABLE_AVRO
+        if (!!conf.schema_registry_url !=
+            !!(conf.flags & (CONF_F_FMT_AVRO_VALUE|CONF_F_FMT_AVRO_KEY)))
+                KC_FATAL("-r requires -s avro and vice-versa");
+
+        if (conf.schema_registry_url) {
+                char *t;
+
+                if (!strchr("GC", conf.mode))
+                        KC_FATAL("Schema-registry support is only available "
+                                 "in the consumer");
+
+                /* Trim trailing slashes from URL to avoid 404 */
+                t = &conf.schema_registry_url[strlen(conf.
+                                                     schema_registry_url)-1];
+                while (t >= conf.schema_registry_url && *t == '/') {
+                        *t = '\0';
+                        t--;
+                }
+
+                if (try_conf_set("schema.registry.url",
+                                 conf.schema_registry_url,
+                                 errstr, sizeof(errstr)) == -1)
+                        KC_FATAL("%s", errstr);
+
+
+                /* Initialize Avro/Schema-Registry client */
+                kc_avro_init(NULL, NULL, NULL, NULL);
+        }
+#endif
+
 
         if (strchr("GC", conf.mode)) {
                 /* Must be explicitly enabled for librdkafka >= v1.0.0 */
@@ -1704,6 +1964,17 @@ int main (int argc, char **argv) {
 	struct timeval tv;
         rd_kafka_topic_partition_list_t *rktparlist = NULL;
 
+        /* Certain Docker images don't have kafkacat as the entry point,
+         * requiring `kafkacat` to be the first argument. As these images
+         * are fixed the examples get outdated and that first argument
+         * will still be passed to the container and thus kafkacat,
+         * so remove it here. */
+        if (argc > 1 && !strcmp(argv[1], "kafkacat")) {
+                if (argc > 2)
+                        memmove(&argv[1], &argv[2], sizeof(*argv) * (argc - 2));
+                argc--;
+        }
+
         signal(SIGINT, term);
         signal(SIGTERM, term);
 #ifdef SIGPIPE
@@ -1794,6 +2065,10 @@ int main (int argc, char **argv) {
 
         rd_kafka_wait_destroyed(5000);
 
+#if ENABLE_AVRO
+        kc_avro_term();
+#endif
+
         fmt_term();
 
         exit(conf.exitcode);
